(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.0 - 7.3)
 *)

Require Export Arith.
Require Import mystuff.

(***************************************************************************)
(* Definition and lemmas of registers in TIS *)
(***************************************************************************)

(* This file must be compiled using coqc to tisreg.vo because it is 
   required as such by tis.v, ie. type in the shell:
coqc tisreg.v
 *)

Inductive _Reg : Set :=
  | _r0 : _Reg
  | _r1 : _Reg
  | _r2 : _Reg
  | _r3 : _Reg
  | _r4 : _Reg
  | _r5 : _Reg
  | _r6 : _Reg
  | _r7 : _Reg
  | _r8 : _Reg
  | _r9 : _Reg
  | _r10 : _Reg
  | _r11 : _Reg
  | _r12 : _Reg
  | _r13 : _Reg
  | _r14 : _Reg
  | _r15 : _Reg
  | _r16 : _Reg
  | _r17 : _Reg
  | _r18 : _Reg
  | _r19 : _Reg
  | _r20 : _Reg
  | _r21 : _Reg
  | _r22 : _Reg
  | _r23 : _Reg
  | _r24 : _Reg
  | _r25 : _Reg
  | _r26 : _Reg
  | _r27 : _Reg
  | _r28 : _Reg
  | _r29 : _Reg
  | _r30 : _Reg
  | _r31 : _Reg.

Definition nat2_reg (n : nat) :=
  match n with
  | O => _r0
  | S O => _r1
  | S (S O) => _r2
  | S (S (S O)) => _r3
  | S (S (S (S O))) => _r4
  | S (S (S (S (S O)))) => _r5
  | S (S (S (S (S (S O))))) => _r6
  | S (S (S (S (S (S (S O)))))) => _r7
  | S (S (S (S (S (S (S (S O))))))) => _r8
  | S (S (S (S (S (S (S (S (S O)))))))) => _r9
  | S (S (S (S (S (S (S (S (S (S O))))))))) => _r10
  | S (S (S (S (S (S (S (S (S (S (S O)))))))))) => _r11
  | S (S (S (S (S (S (S (S (S (S (S (S O))))))))))) => _r12
  | S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))) => _r13
  | S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))) => _r14
  | S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))) => _r15
  | S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))) => _r16
  | S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))))) =>
      _r17
  | S
    (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))))) =>
      _r18
  | S
    (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))))))) =>
      _r19
  | S
    (S
     (S
      (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))))))) =>
      _r20
  | S
    (S
     (S
      (S
       (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))))))))) =>
      _r21
  | S
    (S
     (S
      (S
       (S
        (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))))))))) =>
      _r22
  | S
    (S
     (S
      (S
       (S
        (S
         (S
          (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))))))))))) =>
      _r23
  | S
    (S
     (S
      (S
       (S
        (S
         (S
          (S
           (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))))))))))) =>
      _r24
  | S
    (S
     (S
      (S
       (S
        (S
         (S
          (S
           (S
            (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))))))))))))) =>
      _r25
  | S
    (S
     (S
      (S
       (S
        (S
         (S
          (S
           (S
            (S
             (S
              (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))))))))))))) =>
      _r26
  | S
    (S
     (S
      (S
       (S
        (S
         (S
          (S
           (S
            (S
             (S
              (S
               (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))))))))))))))) =>
      _r27
  | S
    (S
     (S
      (S
       (S
        (S
         (S
          (S
           (S
            (S
             (S
              (S
               (S
                (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))))))))))))))) =>
      _r28
  | S
    (S
     (S
      (S
       (S
        (S
         (S
          (S
           (S
            (S
             (S
              (S
               (S
                (S
                 (S
                  (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))))))))))))))))) =>
      _r29
  | S
    (S
     (S
      (S
       (S
        (S
         (S
          (S
           (S
            (S
             (S
              (S
               (S
                (S
                 (S
                  (S
                   (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))))))))))))))))) =>
      _r30
  | S
    (S
     (S
      (S
       (S
        (S
         (S
          (S
           (S
            (S
             (S
              (S
               (S
                (S
                 (S
                  (S
                   (S
                    (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))))))))))))))))))) =>
      _r31
  | _ => _r0
  end.



Definition beq_reg : forall r1 r2 : _Reg, bool.
intros; assert ({r1 = r2} + {r1 <> r2});
 [ intros; compare r1 r2; auto
 | intros; case H; [ intros; exact true | intros; exact false ] ].
Defined.


Definition _reg2nat (r : _Reg) :=
  let fix getreg (n : nat) : nat :=
         match n with
         | O => 0
         | S m => if beq_reg (nat2_reg n) r then n else getreg m
         end in
  getreg 16.


Lemma beq_reg_neq_false : forall r r', r <> r' -> beq_reg r r' = false.
  intros r r'; case r; case r'; intros; compute in |- *; auto; try AbsurdEq. Qed.
Lemma beq_reg_neq_falseR : forall r r', r' <> r -> beq_reg r r' = false.
  intros; apply beq_reg_neq_false; auto. Qed.
Lemma beq_reg_false_neq : forall r r', r <> r' -> false = beq_reg r r'.
  intros; symmetry  in |- *; apply beq_reg_neq_false; auto. Qed.
Lemma beq_reg_false_neqR : forall r r', r' <> r -> false = beq_reg r r'.
  intros; symmetry  in |- *; apply beq_reg_neq_false; auto. Qed.
Lemma beq_reg_id_true : forall r, beq_reg r r = true.
  intros; induction  r as [| | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | ]; auto. Qed.
Lemma beq_reg_true_id : forall r, true = beq_reg r r.
  intros; induction  r as [| | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | ]; auto. Qed.
Lemma beq_reg_eq_true : forall r r', r = r' -> beq_reg r r' = true.
  intros r r'; case r; case r'; intros; compute in |- *; auto;
   try inversion H. Qed.
Lemma beq_reg_eq_trueR : forall r r', r' = r -> beq_reg r r' = true.
  intros; apply beq_reg_eq_true; auto. Qed.
Lemma beq_reg_true_eq : forall r r', r = r' -> true = beq_reg r r'.
  intros; symmetry  in |- *; apply beq_reg_eq_true; auto. Qed.
Lemma beq_reg_true_eqR : forall r r', r' = r -> true = beq_reg r r'.
  intros; symmetry  in |- *; apply beq_reg_eq_true; auto. Qed.

Hint Immediate beq_reg_neq_false beq_reg_neq_falseR beq_reg_false_neq
  beq_reg_false_neqR beq_reg_id_true beq_reg_true_id beq_reg_eq_true
  beq_reg_eq_trueR beq_reg_true_eq beq_reg_true_eqR.

Lemma eqorneq_tisreg : forall r r' : _Reg, r = r' \/ r <> r'.
intros; compare r r'; auto. Qed.

Hint Immediate eqorneq_tisreg.
