(*
 * Coq code for the paper
 *
 *       A Syntactic Approach to Foundational Proof-Carrying Code
 *                            
 *
 *   by Zhong Shao, Valery Trifonov, Stefan Monnier, Nadeem A. Hamid, 
 *                          and Zhaozhong Ni
 *
 * (for Coq version 7.3)
 *)


(***************************************************************************)
(* Utility lemmas for translation from FTAL to machine state *)
(***************************************************************************)

Require Import talreg.
Require Import tisreg.

(* register translation *)
Definition TrRegF (regi : Reg) : _Reg :=
  match regi with
  | r0 => _r0
  | r1 => _r1
  | r2 => _r2
  | r3 => _r3
  | r4 => _r4
  | r5 => _r5
  | r6 => _r6
  | r7 => _r7
  | r8 => _r8
  | r9 => _r9
  | r10 => _r10
  | r11 => _r11
  | r12 => _r12
  | r13 => _r13
  | r14 => _r14
  | r15 => _r15
  | r16 => _r16
  | r17 => _r17
  | r18 => _r18
  | r19 => _r19
  | r20 => _r20
  | r21 => _r21
  | r22 => _r22
  | r23 => _r23
  | r24 => _r24
  | r25 => _r25
  | r26 => _r26
  | r27 => _r27
  | r28 => _r28
  | r29 => _r29
  | r30 => _r30
  end.


Definition InvTrRegF (r : _Reg) (H : r <> _r31) : Reg :=
  _Reg_rec (fun a : _Reg => a <> _r31 -> Reg) (fun _ : _r0 <> _r31 => r0)
    (fun _ : _r1 <> _r31 => r1) (fun _ : _r2 <> _r31 => r2)
    (fun _ : _r3 <> _r31 => r3) (fun _ : _r4 <> _r31 => r4)
    (fun _ : _r5 <> _r31 => r5) (fun _ : _r6 <> _r31 => r6)
    (fun _ : _r7 <> _r31 => r7) (fun _ : _r8 <> _r31 => r8)
    (fun _ : _r9 <> _r31 => r9) (fun _ : _r10 <> _r31 => r10)
    (fun _ : _r11 <> _r31 => r11) (fun _ : _r12 <> _r31 => r12)
    (fun _ : _r13 <> _r31 => r13) (fun _ : _r14 <> _r31 => r14)
    (fun _ : _r15 <> _r31 => r15) (fun _ : _r16 <> _r31 => r16)
    (fun _ : _r17 <> _r31 => r17) (fun _ : _r18 <> _r31 => r18)
    (fun _ : _r19 <> _r31 => r19) (fun _ : _r20 <> _r31 => r20)
    (fun _ : _r21 <> _r31 => r21) (fun _ : _r22 <> _r31 => r22)
    (fun _ : _r23 <> _r31 => r23) (fun _ : _r24 <> _r31 => r24)
    (fun _ : _r25 <> _r31 => r25) (fun _ : _r26 <> _r31 => r26)
    (fun _ : _r27 <> _r31 => r27) (fun _ : _r28 <> _r31 => r28)
    (fun _ : _r29 <> _r31 => r29) (fun _ : _r30 <> _r31 => r30)
    (fun H0 : _r31 <> _r31 =>
     False_rec Reg
       (let H1 := H0 in
        (let H2 := fun _ : _r31 = _r31 -> False => H0 (refl_equal _r31) in
         fun H3 : _r31 <> _r31 => H2 H3) H1)) r H.



(* lemmas *)

Lemma trreg_neq_31 :
 forall (r : Reg) (_r : _Reg), TrRegF r = _r -> _r <> _r31.
Proof.
intros.
induction r.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
compute in H. rewrite <- H. discriminate.
Qed.

Lemma trreg_inv :
 forall (r : Reg) (_r : _Reg) (D : _r <> _r31),
 TrRegF r = _r -> r = InvTrRegF _r D.
Proof.
(*Double Induction 1 2.*)
simple induction r; simple induction _r.
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. intros; inversion H. 
intros. absurd (_r31 <> _r31); auto.
intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. intros; inversion H. 
intros; compute in |- *; auto.
intros; inversion H. 
Qed.



Lemma trregf_eq_eq :
 forall (r1 r2 : Reg) (r : _Reg), TrRegF r1 = r -> TrRegF r2 = r -> r1 = r2.
simple induction r.
intros. cut (_r0 <> _r31). intros. cut (r1 = InvTrRegF _r0 H1); intros. cut (r2 = InvTrRegF _r0 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r1 <> _r31). intros. cut (r1 = InvTrRegF _r1 H1); intros. cut (r2 = InvTrRegF _r1 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r2 <> _r31). intros. cut (r1 = InvTrRegF _r2 H1); intros. cut (r2 = InvTrRegF _r2 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r3 <> _r31). intros. cut (r1 = InvTrRegF _r3 H1); intros. cut (r2 = InvTrRegF _r3 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r4 <> _r31). intros. cut (r1 = InvTrRegF _r4 H1); intros. cut (r2 = InvTrRegF _r4 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r5 <> _r31). intros. cut (r1 = InvTrRegF _r5 H1); intros. cut (r2 = InvTrRegF _r5 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r6 <> _r31). intros. cut (r1 = InvTrRegF _r6 H1); intros. cut (r2 = InvTrRegF _r6 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r7 <> _r31). intros. cut (r1 = InvTrRegF _r7 H1); intros. cut (r2 = InvTrRegF _r7 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r8 <> _r31). intros. cut (r1 = InvTrRegF _r8 H1); intros. cut (r2 = InvTrRegF _r8 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r9 <> _r31). intros. cut (r1 = InvTrRegF _r9 H1); intros. cut (r2 = InvTrRegF _r9 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r10 <> _r31). intros. cut (r1 = InvTrRegF _r10 H1); intros. cut (r2 = InvTrRegF _r10 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r11 <> _r31). intros. cut (r1 = InvTrRegF _r11 H1); intros. cut (r2 = InvTrRegF _r11 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r12 <> _r31). intros. cut (r1 = InvTrRegF _r12 H1); intros. cut (r2 = InvTrRegF _r12 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r13 <> _r31). intros. cut (r1 = InvTrRegF _r13 H1); intros. cut (r2 = InvTrRegF _r13 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r14 <> _r31). intros. cut (r1 = InvTrRegF _r14 H1); intros. cut (r2 = InvTrRegF _r14 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r15 <> _r31). intros. cut (r1 = InvTrRegF _r15 H1); intros. cut (r2 = InvTrRegF _r15 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r16 <> _r31). intros. cut (r1 = InvTrRegF _r16 H1); intros. cut (r2 = InvTrRegF _r16 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r17 <> _r31). intros. cut (r1 = InvTrRegF _r17 H1); intros. cut (r2 = InvTrRegF _r17 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r18 <> _r31). intros. cut (r1 = InvTrRegF _r18 H1); intros. cut (r2 = InvTrRegF _r18 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r19 <> _r31). intros. cut (r1 = InvTrRegF _r19 H1); intros. cut (r2 = InvTrRegF _r19 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r20 <> _r31). intros. cut (r1 = InvTrRegF _r20 H1); intros. cut (r2 = InvTrRegF _r20 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r21 <> _r31). intros. cut (r1 = InvTrRegF _r21 H1); intros. cut (r2 = InvTrRegF _r21 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r22 <> _r31). intros. cut (r1 = InvTrRegF _r22 H1); intros. cut (r2 = InvTrRegF _r22 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r23 <> _r31). intros. cut (r1 = InvTrRegF _r23 H1); intros. cut (r2 = InvTrRegF _r23 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r24 <> _r31). intros. cut (r1 = InvTrRegF _r24 H1); intros. cut (r2 = InvTrRegF _r24 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r25 <> _r31). intros. cut (r1 = InvTrRegF _r25 H1); intros. cut (r2 = InvTrRegF _r25 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r26 <> _r31). intros. cut (r1 = InvTrRegF _r26 H1); intros. cut (r2 = InvTrRegF _r26 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r27 <> _r31). intros. cut (r1 = InvTrRegF _r27 H1); intros. cut (r2 = InvTrRegF _r27 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r28 <> _r31). intros. cut (r1 = InvTrRegF _r28 H1); intros. cut (r2 = InvTrRegF _r28 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r29 <> _r31). intros. cut (r1 = InvTrRegF _r29 H1); intros. cut (r2 = InvTrRegF _r29 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r30 <> _r31). intros. cut (r1 = InvTrRegF _r30 H1); intros. cut (r2 = InvTrRegF _r30 H1); intros. compute in H2. compute in H3. rewrite H2; rewrite H3; auto. apply trreg_inv; auto. apply trreg_inv; auto. apply trreg_neq_31 with r1; auto. 
intros. cut (_r31 <> _r31). intros. absurd (_r31 <> _r31); auto. apply trreg_neq_31 with r1; auto. 
Qed.


Lemma trregf_neq_neq :
 forall (r1 r2 : Reg) (r r' : _Reg),
 TrRegF r1 = r -> TrRegF r2 = r' -> r <> r' -> r1 <> r2. 
Proof.
intros.
cut (r1 = r2 \/ r1 <> r2); auto.
intros.
inversion_clear H2.
rewrite H3 in H.
cut (r = r').
intros.
rewrite H2 in H1.
absurd (r' <> r').
auto.
auto.
rewrite H in H0.
auto.
auto.
apply eqorneq_reg.
Qed.